#include 								<Arduino.h>

/*

This program monitors a motion detector for a PC.

It receives commands in the form

Position	Length		Field Name		Field Function
0			1			Checksum		XOR of all bytes after this
1			2			RecLen			Total length of record
3			1			Command			One of:
										0 - No command
										1 - Is Switch On?
										2 - Is motion detected?
										3 - Turn LED on
										4 - Turn LED off
										5 - Alive Packet (sent by Arduino only)
										9 - Error (sent from arduino to pc)
4			?			Response		Command		Response
										1			T or F
										2			T or F
										6			string

The PC is sends the commands. When the arduino receives a command, it will reply with
the same command with additional information in the response field.

*/

//	Arduino2PCviaSerial
//		1.0		Nov 2014				Dan Hallock

// -----------------------------------------------------------------------------
// constants

const int								externalLEDPin		= 	3;
const int								internalLEDPin		= 	13;
const int								maxPacketSize		= 	20;				// max size of buffers in bytes
const int 								pirPin				=	9;
const int								swPin				=   4;
const unsigned long						timeoutOffset		=	10L * 1000L; 	// when to indicate motion has stopped

// -----------------------------------------------------------------------------
// externals

// a2p_ are the primary routines to communicate with the pc - get and send strings
boolean			a2p_getStr				(char* s);
void 			a2p_sendStr				(const char* s);

// a2p_i_ are routines called by a2p_ routines
unsigned char 	a2p_i_computeChecksum	(char* packet);
boolean			a2p_i_decodePacket		(char* packet, char* s);
void 			a2p_i_encodePacket		(const char* s, char* packet);
int 			a2p_i_getPacketLen		(char* packet);
boolean 		a2p_i_recvPacket		(char* packet, int packetMax);
void 			a2p_i_xmitPacket		(char* packet);

// misc routines used by loop
char*			getCommand				(char* s);
boolean 		motionDetected			();

// -----------------------------------------------------------------------------
// globals

unsigned long							aliveTimeout		= 0L;				// when 'alive' times out and onboard led turns off

// -----------------------------------------------------------------------------
// get a string from the PC. This will be normal \0 terminated C string. If an error
// occurs decoding string, return false. If there is nothing read, we just return
// null string.
boolean  a2p_getStr(
	char* 								s
	) {
	char								packet[maxPacketSize];

	// get the next packet from the pc
	if (a2p_i_recvPacket(packet,sizeof(packet)) == false) {
		// if nothing received, return null string
		strcpy(s, "");
		return true;
		}

	// decode packet. if any error, return error
    if (!a2p_i_decodePacket(packet, s)) {					// if decoding problem return error code
		strcpy(s, "");
		return false;
		}

	return true;

	} // a2p_sendStr

// -----------------------------------------------------------------------------
void a2p_sendStr(
	const char* 						s
	) {
	char								packet[maxPacketSize];

    a2p_i_encodePacket(s, packet);
    a2p_i_xmitPacket(packet);

	} // a2p_sendStr

// -----------------------------------------------------------------------------
unsigned char a2p_i_computeChecksum(
	char* 								packet
	) {
	unsigned char						cksum;
	int									k;

    cksum = 0;

	// compute the basic checksum by xor'ing every byte in the packet EXCEPT the checksum
    for (k = 1; k <= a2p_i_getPacketLen(packet); k = k + 1) {
        cksum = cksum xor packet[k];
        }

	// Clear high order bit as it might have trouble being transmitted.
	// also turn any control characters into a normal character.
    cksum = cksum & 0x7F;
    if (cksum < 0x20)
        cksum = cksum | 0x20;

	return cksum;

	} // a2p_i_computeChecksum

// -----------------------------------------------------------------------------
// Verify the packet is valid, checking len and checksum, then return the payload
// in s.
boolean a2p_i_decodePacket(
    char* 								packet,
    char* 								s
) {
    int									k;
    int									len;
    unsigned char						cksum;
    int									cmd;

	strcpy(s, "");

    if (strlen(packet) < 4) {								// command must be at least 4 bytes
        return false;
        }

    // determine length of packet
    len = a2p_i_getPacketLen(packet);
    if (len < 3 || len > maxPacketSize-1)
		return false;

    cksum = a2p_i_computeChecksum(packet);
    if (cksum != packet[0]) {
		return false;
		}

    // transfer the packet payload to s
    for (k = 3; k < len; k = k + 1) {
        s[k-3] = packet[k];
        }
    s[len-3] = '\0';

    return true;

    } // a2p_i_decodepacket

// -----------------------------------------------------------------------------
// given data, s, append packet header to s and return in packet.
void a2p_i_encodePacket(
	const char* 						s,
	char*  								packet
	) {
	unsigned char						cksum;
	int									k;
	int									len;

	strcpy(packet, "000");
	strncat(packet, s, maxPacketSize);

	len = strlen(packet);
	for (k = 2; k >= 1; k = k - 1) {
		packet[k] = (len % 10) + '0';
		len = len / 10;
		}

	cksum = a2p_i_computeChecksum(packet);
	packet[0] = cksum;

	} // a2p_i_encodePacket

// -----------------------------------------------------------------------------
// return the length of the packet encoded in the packet.
int a2p_i_getPacketLen(
	char* 								packet
	) {
	int									i;
	int									len;

    len = 0;											// initialize length of command
    for (i = 1; i <= 2; i = i + 1) {
        if (!isdigit((packet[i]))) {
            return -1;
            }
        len = len * 10;									// left shift contents of len
        len = len + (packet[i] - '0');						// convert s[i] to int and add to len
        }
	return len;
	} // a2p_i_getPacketLen

// -----------------------------------------------------------------------------
// receive next packet from wire.
boolean a2p_i_recvPacket(
	char* 								packet,
	int									packetMax
	){
    char								c;
    int									i;
    int									t;

    if (!Serial.available())
        return false;

	i = 0;													// init s's index
	t = 0;
    while (true) {
        if (Serial.available()) {							// anything to read?
            if (i > packetMax-1) {							// bounds check
                return false;
                }
            c = Serial.read();								// read next byte
            if (c =='\r') {									// if CR, finished reading command
                packet[i] = '\0';								// terminate string nicely
                break; //while								// terminate while look
                }
            packet[i] = c;										// append byte to string
            i = i + 1;										// incr index
            }
        else { 												// wait a bit for more data
			// if it takes more than 2 secs to get the packet, something is wrong
			t = t + 100;
			if (t > (2 * 10000) / 100) {
				strcpy(packet, "");
				return false;
				}
            delay(100L);
            }
        } // while

	return true;

	} // a2p_i_recvPacket

// -----------------------------------------------------------------------------
// Transmit the packet out on the line.
void a2p_i_xmitPacket(
	char* 								packet
	) {
	int									i;
	int									len;

	len = a2p_i_getPacketLen(packet);

	for (i = 0; i <= len-1; i = i + 1) {
		Serial.print(packet[i]);
		}

	Serial.println();

	} // a2p_i_xmitPacket

// -----------------------------------------------------------------------------
// Get the next command from the PC. If any issue, put "9" in the string as the
// command indicating there was a problem.
char* getCommand(
    char*								s
) {
    char								packet[maxPacketSize];

	// if decoding problem return error code
    if (!a2p_getStr(s))
		strcpy(s, "9");

	return s;
    } // getCommand

// -----------------------------------------------------------------------------
// Return true if any motion has been detected.
boolean motionDetected(
) {
    static unsigned long				timeout 			= 0L;

    if (digitalRead(pirPin) == HIGH) {
        if (timeout == 0) { 								// first motion?
            delay(100L);									// wait a bit for debounce
            timeout = millis() + timeoutOffset;
            }
        else {												// motion continues, extend timeout
            timeout = millis() + timeoutOffset;
            }
        }
    else {
        if (timeout == 0) {									// no movement and timed out, do nothing
            }
        else if (millis() > timeout) {						// no movement, and timing out
            timeout = 0L;
            }
        else {												// no movement, still waiting to timeout
            }
        }
    if (timeout == 0L)
        return false;
    else
        return true;

    } // motionDetected

// -----------------------------------------------------------------------------
void setup(
	) {
    Serial.begin(9600);

    pinMode(pirPin, INPUT);
    pinMode(swPin, 	INPUT);
    digitalWrite(swPin, HIGH);								// set pull-up resistor

    pinMode(internalLEDPin, OUTPUT);
    pinMode(externalLEDPin, OUTPUT);

    digitalWrite(internalLEDPin, HIGH);
    digitalWrite(externalLEDPin, HIGH);

    delay(1L * 1000L);										// give time to calibrate motion detector

    digitalWrite(internalLEDPin, LOW);
    digitalWrite(externalLEDPin, LOW);

    } // setup

// -----------------------------------------------------------------------------
void loop(
	) {
    int									cmd;
    char								s[maxPacketSize];

    while (true) {
        motionDetected();									// always check motion detector for activity
        if (millis() > aliveTimeout) {
			digitalWrite(internalLEDPin, LOW);
			}

        getCommand(s);
        if (strlen(s) == 0)
			continue;

        switch (s[0]) {
			case '0':											// no command received
				break;
            case '1': 										// is SW on?
				if (digitalRead(swPin) == HIGH)
					a2p_sendStr("1F");
				else
					a2p_sendStr("1T");
                break;
            case '2':											// is motion detected?
				if (motionDetected())
					a2p_sendStr("2T");
				else
					a2p_sendStr("2F");
                break;
            case '3': 										// turn LED on
				digitalWrite(externalLEDPin, HIGH);
				a2p_sendStr("3");
                break;
            case '4':											// turn LED off
				digitalWrite(externalLEDPin, LOW);
				a2p_sendStr("4");
                break;
            case '5':											// alive packet
				digitalWrite(internalLEDPin, HIGH);
				aliveTimeout = millis() + (10L * 1000L);	// when to turn alive LED out again
				a2p_sendStr("5");
                break;
            case '9':											// handle error in cmd received
                a2p_sendStr("9");
                break;
            default:										// unexpected command
				a2p_sendStr("9");
                break;
            } // switch
        } // while
    } // loop
